@@ -46,6 +46,7 @@ TITLE_FIDS         = [TITLE_FID, SUBTITLE_FID, CONTENT_TITLE_FID];
 COMMENT_FID        = "COMM";
 GENRE_FID          = "TCON";
 TRACKNUM_FID       = "TRCK";
+DISCNUM_FID        = "TPOS";
 USERTEXT_FID       = "TXXX";
 CDID_FID           = "MCDI";
 IMAGE_FID          = "APIC";
@@ -64,6 +65,7 @@ URL_FIDS           = [URL_COMMERCIAL_FID, URL_COPYRIGHT_FID,
 USERURL_FID        = "WXXX";
 PLAYCOUNT_FID      = "PCNT";
 UNIQUE_FILE_ID_FID = "UFID";
+BPM_FID            = "TBPM";
 
 obsoleteFrames = {"EQUA": "Equalisation",
                   "IPLS": "Involved people list",
@@ -80,6 +82,9 @@ OBSOLETE_TIME_FID            = "TIME";
 OBSOLETE_ORIG_RELEASE_FID    = "TORY";
 OBSOLETE_RECORDING_DATE_FID  = "TRDA";
 
+DATE_FIDS          = ["TDRL", "TDOR", "TDRC", OBSOLETE_YEAR_FID,
+                      OBSOLETE_DATE_FID];
+
 frameDesc = { "AENC": "Audio encryption",
               "APIC": "Attached picture",
               "ASPI": "Audio seek point index",
@@ -176,6 +181,78 @@ frameDesc = { "AENC": "Audio encryption",
               "WPUB": "Publishers official webpage",
               "WXXX": "User defined URL link frame" };
 
+
+# mapping of 2.2 frames to 2.3/2.4
+TAGS2_2_TO_TAGS_2_3_AND_4 = {                  
+    "TT1" : "TIT1", # CONTENTGROUP content group description
+    "TT2" : "TIT2", # TITLE title/songname/content description
+    "TT3" : "TIT3", # SUBTITLE subtitle/description refinement
+    "TP1" : "TPE1", # ARTIST lead performer(s)/soloist(s)
+    "TP2" : "TPE2", # BAND band/orchestra/accompaniment
+    "TP3" : "TPE3", # CONDUCTOR conductor/performer refinement
+    "TP4" : "TPE4", # MIXARTIST interpreted, remixed, modified by
+    "TCM" : "TCOM", # COMPOSER composer
+    "TXT" : "TEXT", # LYRICIST lyricist/text writer
+    "TLA" : "TLAN", # LANGUAGE language(s)
+    "TCO" : "TCON", # CONTENTTYPE content type
+    "TAL" : "TALB", # ALBUM album/movie/show title
+    "TRK" : "TRCK", # TRACKNUM track number/position in set
+    "TPA" : "TPOS", # PARTINSET part of set
+    "TRC" : "TSRC", # ISRC international standard recording code
+    "TDA" : "TDAT", # DATE date
+    "TYE" : "TYER", # YEAR year
+    "TIM" : "TIME", # TIME time
+    "TRD" : "TRDA", # RECORDINGDATES recording dates
+    "TOR" : "TORY", # ORIGYEAR original release year
+    "TBP" : "TBPM", # BPM beats per minute
+    "TMT" : "TMED", # MEDIATYPE media type
+    "TFT" : "TFLT", # FILETYPE file type
+    "TCR" : "TCOP", # COPYRIGHT copyright message
+    "TPB" : "TPUB", # PUBLISHER publisher
+    "TEN" : "TENC", # ENCODEDBY encoded by
+    "TSS" : "TSSE", # ENCODERSETTINGS software/hardware + settings for encoding
+    "TLE" : "TLEN", # SONGLEN length (ms)
+    "TSI" : "TSIZ", # SIZE size (bytes)
+    "TDY" : "TDLY", # PLAYLISTDELAY playlist delay
+    "TKE" : "TKEY", # INITIALKEY initial key
+    "TOT" : "TOAL", # ORIGALBUM original album/movie/show title
+    "TOF" : "TOFN", # ORIGFILENAME original filename
+    "TOA" : "TOPE", # ORIGARTIST original artist(s)/performer(s)
+    "TOL" : "TOLY", # ORIGLYRICIST original lyricist(s)/text writer(s)
+    "TXX" : "TXXX", # USERTEXT user defined text information frame
+    "WAF" : "WOAF", # WWWAUDIOFILE official audio file webpage
+    "WAR" : "WOAR", # WWWARTIST official artist/performer webpage
+    "WAS" : "WOAS", # WWWAUDIOSOURCE official audion source webpage
+    "WCM" : "WCOM", # WWWCOMMERCIALINFO commercial information
+    "WCP" : "WCOP", # WWWCOPYRIGHT copyright/legal information
+    "WPB" : "WPUB", # WWWPUBLISHER publishers official webpage
+    "WXX" : "WXXX", # WWWUSER user defined URL link frame
+    "IPL" : "IPLS", # INVOLVEDPEOPLE involved people list
+    "ULT" : "USLT", # UNSYNCEDLYRICS unsynchronised lyrics/text transcription
+    "COM" : "COMM", # COMMENT comments
+    "UFI" : "UFID", # UNIQUEFILEID unique file identifier
+    "MCI" : "MCDI", # CDID music CD identifier
+    "ETC" : "ETCO", # EVENTTIMING event timing codes
+    "MLL" : "MLLT", # MPEGLOOKUP MPEG location lookup table
+    "STC" : "SYTC", # SYNCEDTEMPO synchronised tempo codes
+    "SLT" : "SYLT", # SYNCEDLYRICS synchronised lyrics/text
+    "RVA" : "RVAD", # VOLUMEADJ relative volume adjustment
+    "EQU" : "EQUA", # EQUALIZATION equalization
+    "REV" : "RVRB", # REVERB reverb
+    "PIC" : "APIC", # PICTURE attached picture
+    "GEO" : "GEOB", # GENERALOBJECT general encapsulated object
+    "CNT" : "PCNT", # PLAYCOUNTER play counter
+    "POP" : "POPM", # POPULARIMETER popularimeter
+    "BUF" : "RBUF", # BUFFERSIZE recommended buffer size
+    "CRA" : "AENC", # AUDIOCRYPTO audio encryption
+    "LNK" : "LINK", # LINKEDINFO linked information
+    # Extension workarounds i.e., ignore them
+    "TCP" : "TCP ", # iTunes "extension" for compilation marking
+    "CM1" : "CM1 "  # Seems to be some script kiddie tagging the tag.
+                    # For example, [rH] join #rH on efnet [rH]
+}
+
+
 NULL_FRAME_FLAGS = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
 
 TEXT_FRAME_RX = re.compile("^T[A-Z0-9][A-Z0-9][A-Z0-9]$");
@@ -199,7 +276,9 @@ KNOWN_BAD_FRAMES = [
     "\x00MP",
     " MP",
     "MP3",
-    "COM "
+    "COM ",
+    "TCP ", # iTunes
+    "CM1 "  # Script kiddie
 ]
 
 LATIN1_ENCODING   = "\x00";
@@ -212,6 +291,9 @@ DEFAULT_ID3_MAJOR_VERSION = 2;
 DEFAULT_ID3_MINOR_VERSION = 4;
 DEFAULT_LANG = "eng";
 
+def cleanNulls(s):
+   return "/".join([x for x in s.split('\x00') if x])
+
 def id3EncodingToString(encoding):
     if encoding == LATIN1_ENCODING:
         return "latin_1";
@@ -225,14 +307,9 @@ def id3EncodingToString(encoding):
         raise ValueError;
 
 ################################################################################
-class FrameException:
-   msg = "";
-
-   def __init__(self, msg):
-      self.msg = msg;
-
-   def __str__(self):
-      return self.msg;
+class FrameException(Exception):
+    '''Thrown by invalid frames'''
+    pass;
 
 ################################################################################
 class FrameHeader:
@@ -291,7 +368,10 @@ class FrameHeader:
 
       # 1.x tags are converted to 2.4 frames internally.  These frames are
       # created with frame flags \x00.
-      if (major == 2 and minor == 3):
+      if (major == 2 and minor == 2):
+          # no flags for 2.2 frames
+          pass;
+      elif (major == 2 and minor == 3):
          self.TAG_ALTER   = 0;   
          self.FILE_ALTER  = 1;   
          self.READ_ONLY   = 2;   
@@ -342,12 +422,44 @@ class FrameHeader:
 
       return data;
 
+   def parse2_2(self, f):
+      frameId = f.read(3);
+      frameId = map2_2FrameId(frameId);
+      if self.isFrameIdValid(frameId):
+         TRACE_MSG("FrameHeader [id]: %s (0x%x%x%x)" % (frameId,
+                                                       ord(frameId[0]),
+                                                       ord(frameId[1]),
+                                                       ord(frameId[2])));
+         self.id = frameId;
+         # dataSize corresponds to the size of the data segment after
+         # encryption, compression, and unsynchronization.
+         sz = f.read(3);
+         self.dataSize = bin2dec(bytes2bin(sz, 8));
+         TRACE_MSG("FrameHeader [data size]: %d (0x%X)" % (self.dataSize,
+                                                           self.dataSize));
+      elif frameId == '\x00\x00\x00':
+         TRACE_MSG("FrameHeader: Null frame id found at byte " +\
+                   str(f.tell()));
+         return 0;
+      elif not strictID3() and frameId in KNOWN_BAD_FRAMES:
+         TRACE_MSG("FrameHeader: Illegal but known "\
+                   "(possibly created by the shitty mp3ext) frame found; "\
+                   "Happily ignoring!" + str(f.tell()));
+         return 0;
+      else:
+         raise FrameException("FrameHeader: Illegal Frame ID: " + frameId);
+      return 1;
+      
+
    # Returns 1 on success and 0 when a null tag (marking the beginning of
    # padding).  In the case of an invalid frame header, a FrameException is 
    # thrown.
    def parse(self, f):
       TRACE_MSG("FrameHeader [start byte]: %d (0x%X)" % (f.tell(),
                                                          f.tell()));
+      if self.minorVersion == 2:
+          return self.parse2_2(f)
+      
       frameId = f.read(4);
       if self.isFrameIdValid(frameId):
          TRACE_MSG("FrameHeader [id]: %s (0x%x%x%x%x)" % (frameId,
@@ -441,7 +553,7 @@ class Frame:
        assert(isinstance(frameHeader, FrameHeader));
        self.header = frameHeader;
 
-   def __repr__(self):
+   def __str__(self):
       desc = self.getFrameDesc();
       return '<%s Frame (%s)>' % (desc, self.header.id);
 
@@ -554,6 +666,13 @@ class Frame:
          except KeyError:
             return "UNKOWN FRAME";
 
+   def getTextDelim(self):
+       if self.encoding == UTF_16_ENCODING or \
+          self.encoding == UTF_16BE_ENCODING:
+           return "\x00\x00";
+       else:
+           return "\x00";
+
 ################################################################################
 class TextFrame(Frame):
    text = u"";
@@ -583,16 +702,21 @@ class TextFrame(Frame):
       TRACE_MSG("TextFrame encoding: %s" % id3EncodingToString(self.encoding));
       try:
           self.text = unicode(data[1:], id3EncodingToString(self.encoding));
+          if not strictID3():
+              self.text = cleanNulls(self.text)
       except TypeError, excArg:
           # if data is already unicode, just copy it
           if excArg.args == ("decoding Unicode is not supported",):
               self.text = data[1:];
+              if not strictID3():
+                 self.text = cleanNulls(self.text)
           else:
               raise;
       TRACE_MSG("TextFrame text: %s" % self.text);
 
-   def __repr__(self):
-      return '<%s (%s): %s>' % (self.getFrameDesc(), self.header.id, self.text);
+   def __unicode__(self):
+      return u'<%s (%s): %s>' % (self.getFrameDesc(), self.header.id,
+                                 self.text);
    
    def render(self):
        if self.header.minorVersion == 4 and self.header.id == "TSIZ":
@@ -621,7 +745,7 @@ class DateFrame(TextFrame):
 
    def _set(self, data, frameHeader):
       TextFrame._set(self, data, frameHeader);
-      if self.header.id[:2] != "TD" and self.header.minorVersion != 3:
+      if self.header.id[:2] != "TD" and self.header.minorVersion >= 4:
          raise FrameException("Invalid frame id for DateFrame: " + \
                               self.header.id);
    
@@ -638,14 +762,8 @@ class DateFrame(TextFrame):
                self.date = d;
             else:
                assert(isinstance(d, unicode));
-               d = d.strip("\x00");
-
                # Witnessed oddball tags with NULL bytes (ozzy.tag from id3lib)
-               strippedDate = "";
-               for c in d:
-                  if ord(c) != 0:
-                     strippedDate += c;
-               d = strippedDate;
+               d = d.strip("\x00");
 
                try:
                   self.date = time.strptime(d, fmt);
@@ -766,19 +884,21 @@ class UserTextFrame(TextFrame):
       self.description = unicode(d, id3EncodingToString(self.encoding));
       TRACE_MSG("UserTextFrame description: %s" % self.description);
       self.text = unicode(t, id3EncodingToString(self.encoding));
+      if not strictID3():
+          self.text = cleanNulls(self.text)
       TRACE_MSG("UserTextFrame text: %s" % self.text);
 
    def render(self):
       data = self.encoding +\
              self.description.encode(id3EncodingToString(self.encoding)) +\
-             "\x00" +\
+             self.getTextDelim() +\
              self.text.encode(id3EncodingToString(self.encoding));
       return self.assembleFrame(data);
 
-   def __repr__(self):
-       return '<%s (%s): {Desc: %s} %s>' % (self.getFrameDesc(),
-                                           self.header.id,
-                                           self.description, self.text);
+   def __unicode__(self):
+       return u'<%s (%s): {Desc: %s} %s>' % (self.getFrameDesc(),
+                                            self.header.id,
+                                            self.description, self.text);
 
 ################################################################################
 class URLFrame(Frame):
@@ -802,12 +922,14 @@ class URLFrame(Frame):
          raise FrameException("Invalid frame id for URLFrame: " + fid);
       data = self.disassembleFrame(data);
       self.url = data;
+      if not strictID3():
+          self.url = cleanNulls(self.url)
 
    def render(self):
       data = str(self.url);
       return self.assembleFrame(data);
 
-   def __repr__(self):
+   def __str__(self):
       return '<%s (%s): %s>' % (self.getFrameDesc(), self.header.id,
                                 self.url);
 
@@ -846,16 +968,18 @@ class UserURLFrame(URLFrame):
       self.description = unicode(d, id3EncodingToString(self.encoding));
       TRACE_MSG("UserURLFrame description: %s" % self.description);
       self.url = u;
+      if not strictID3():
+          self.url = cleanNulls(self.url)
       TRACE_MSG("UserURLFrame text: %s" % self.url);
 
    def render(self):
       data = self.encoding +\
              self.description.encode(id3EncodingToString(self.encoding)) +\
-             "\x00" + self.url;
+             self.getTextDelim() + self.url;
       return self.assembleFrame(data);
 
-   def __repr__(self):
-      return '<%s (%s): %s [Encoding: %s] [Desc: %s]>' %\
+   def __unicode__(self):
+      return u'<%s (%s): %s [Encoding: %s] [Desc: %s]>' %\
              (self.getFrameDesc(), self.header.id,
               self.url, self.encoding, self.description)
 
@@ -919,6 +1043,9 @@ class CommentFrame(Frame):
           else:
               self.description = u"";
               self.comment = u"";
+      if not strictID3():
+          self.description = cleanNulls(self.description)
+          self.comment = cleanNulls(self.comment)
 
    def render(self):
       lang = self.lang.encode("ascii");
@@ -928,12 +1055,12 @@ class CommentFrame(Frame):
           lang = lang + ('\x00' * (3 - len(lang)));
       data = self.encoding + lang +\
              self.description.encode(id3EncodingToString(self.encoding)) +\
-             "\x00" +\
+             self.getTextDelim() +\
              self.comment.encode(id3EncodingToString(self.encoding));
       return self.assembleFrame(data);
 
-   def __repr__(self):
-      return "<%s (%s): %s [Lang: %s] [Desc: %s]>" %\
+   def __unicode__(self):
+      return u"<%s (%s): %s [Lang: %s] [Desc: %s]>" %\
              (self.getFrameDesc(), self.header.id, self.comment,
               self.lang, self.description);
 
@@ -1070,25 +1197,27 @@ class ImageFrame(Frame):
       TRACE_MSG("APIC picture type: " + str(self.pictureType));
 
       self.desciption = u"";
-      buffer = "";
-      ch = input.read(1);
-      while ch != "\x00":
-         buffer += ch;
-         ch = input.read(1);
-      self.description = unicode(buffer, id3EncodingToString(self.encoding));
+
+      # Remaining data is a NULL separated description and image data
+      buffer = input.read();
+      input.close();
+
+      (desc, img) = splitUnicode(buffer, self.encoding);
+      TRACE_MSG("description len: %d" % len(desc));
+      TRACE_MSG("description len: %d" % len(img));
+      self.description = unicode(desc, id3EncodingToString(self.encoding));
       TRACE_MSG("APIC description: " + self.description);
 
       if self.mimeType.find("-->") != -1:
          self.imageData = None;
-         self.imageURL = input.read();
+         self.imageURL = img;
       else:
-         self.imageData = input.read();
+         self.imageData = img;
          self.imageURL = None;
       TRACE_MSG("APIC image data: " + str(len(self.imageData)) + " bytes");
       if strictID3() and not self.imageData and not self.imageURL:
          raise FrameException("APIC frame does not contain any image data");
 
-      input.close();
 
    def writeFile(self, path = "./", name = None):
       if not self.imageData:
@@ -1110,7 +1239,7 @@ class ImageFrame(Frame):
       data = self.encoding + self.mimeType + "\x00" +\
              bin2bytes(dec2bin(self.pictureType, 8)) +\
              self.description.encode(id3EncodingToString(self.encoding)) +\
-             "\x00";
+             self.getTextDelim();
       if self.imageURL:
           data += self.imageURL.encode("ascii");
       else:
@@ -1436,8 +1565,8 @@ class FrameSet(list):
          imageFrames = self[fid];
          for frm in imageFrames:
             if frm.pictureType == frame.pictureType:
-               raise FrameException("Multiple %s frames with the same\
-                                     content descriptor now allowed." % fid);
+               raise FrameException("Multiple %s frames with the same "\
+                                    "content descriptor now allowed." % fid);
 
       # Play count frame (PCNT).  There may be only one
       if PLAYCOUNT_FRAME_RX.match(fid) and self[fid]:
@@ -1480,7 +1609,10 @@ class FrameSet(list):
           h.id = frameId;
           if not encoding:
               encoding = DEFAULT_ENCODING;
-          self.addFrame(TextFrame(h, encoding = encoding, text = text));
+          if frameId in DATE_FIDS:
+              self.addFrame(DateFrame(h, encoding = encoding, date_str = text));
+          else:
+              self.addFrame(TextFrame(h, encoding = encoding, text = text));
 
    # If a user text frame with the same description exists then
    # the frame text is replaced, otherwise the frame is added.
@@ -1626,10 +1758,8 @@ def splitUnicode(data, encoding):
     elif encoding == UTF_16_ENCODING:
         (d, t) = data.split("\x00\x00\x00", 1);
         d += "\x00";
-        t += "\x00";
         return (d, t);
 
-
 #######################################################################
 # Create and return the appropriate frame.
 # Exceptions: ....
@@ -1674,3 +1804,12 @@ def createFrame(frameHeader, data):
      f = UnknownFrame(frameHeader, data);
 
   return f;
+
+
+def map2_2FrameId(originalId):
+    
+    if not TAGS2_2_TO_TAGS_2_3_AND_4.has_key(originalId): return originalId
+    return TAGS2_2_TO_TAGS_2_3_AND_4[originalId]
+        
+        
+    
