@@ -1,4 +1,4 @@
-################################################################################
+###############################################################################
 #
 #  Copyright (C) 2002-2005  Travis Shirk <travis@pobox.com>
 #
@@ -29,14 +29,8 @@ from binfuncs import *;
 ID3_V1_COMMENT_DESC = "ID3 v1 Comment";
 
 ################################################################################
-class TagException:
-   msg = "";
-
-   def __init__(self, msg):
-      self.msg = msg;
-
-   def __str__(self):
-      return self.msg;
+class TagException(Exception):
+   '''error reading tag'''
 
 ################################################################################
 class TagHeader:
@@ -120,7 +114,7 @@ class TagHeader:
       TRACE_MSG("TagHeader [major]: " + str(major));
       TRACE_MSG("TagHeader [minor]: " + str(minor));
       TRACE_MSG("TagHeader [revis]: " + str(rev));
-      if not (major == 2 and (minor == 3 or minor == 4)):
+      if not (major == 2 and (minor >= 2 and minor <= 4)):
          raise TagException("ID3 v" + str(major) + "." + str(minor) +\
                             " is not supported.");
       # Get all the version madness in sync.
@@ -482,6 +476,8 @@ class Tag:
 
       self.setVersion(version);
       version = self.getVersion();
+      if version == ID3_V2_2:
+          raise TagException("Unable to write ID3 v2.2");
       # If v1.0 is being for explicitly then so be it, if not and there is
       # a track number then bumping to v1.1 is /probably/ best.
       if self.header.majorVersion == 1 and self.header.minorVersion == 0 and\
@@ -596,8 +592,7 @@ class Tag:
 
    def getDate(self, fid = None):
        if not fid:
-           for fid in ["TDRL", "TDOR", "TDRC",
-                       OBSOLETE_YEAR_FID, OBSOLETE_DATE_FID]:
+           for fid in DATE_FIDS:
                if self.frames[fid]:
                    return self.frames[fid];
            return None;
@@ -621,22 +616,29 @@ class Tag:
       else:
          return None;
 
+   def _getNum(self, fid):
+      tn = None
+      tt = None
+      f = self.frames[fid];
+      if f:
+         n = f[0].text.split('/')
+         if len(n) == 1:
+            tn = self.toInt(n[0])
+         elif len(n) == 2:
+            tn = self.toInt(n[0])
+            tt = self.toInt(n[1])
+      return (tn, tt)      
+
    # Returns a tuple with the first value containing the track number and the
    # second the total number of tracks.  One or both of these values may be
    # None depending on what is available in the tag. 
    def getTrackNum(self):
-      f = self.frames[TRACKNUM_FID];
-      if f:
-         n = string.split(f[0].text, '/');
-         tn = self.toInt(n[0]);
-         tt = None;
-         if len(n) == 2:
-            tt = self.toInt(n[1]);
-         if tn != None and tt != None:
-            return (tn, tt);
-         elif tn != None:
-            return (tn, None);
-      return (None, None);
+      return self._getNum(TRACKNUM_FID)
+
+   # Like TrackNum, except for DiscNum--that is, position in a set. Most
+   # songs won't have this or it will be 1/1.
+   def getDiscNum(self):
+      return self._getNum(DISCNUM_FID)
 
    # Since multiple comment frames are allowed this returns a list with 0
    # or more elements.  The elements are not the comment strings, they are
@@ -694,8 +696,9 @@ class Tag:
    def setDate(self, year, month = None, dayOfMonth = None,
                hour = None, minute = None, second = None, fid = None):
       if not year and not fid:
-          dateFrame = self.getDate;
-          self.frames.removeFramesByID(dateFrame.id);
+          dateFrames = self.getDate();
+          if dateFrames:
+              self.frames.removeFramesByID(dateFrames[0].header.id);
           return;
       elif not year:
           self.frames.removeFramesByID(fid);
@@ -758,8 +761,14 @@ class Tag:
    # second the total number of tracks.  One or both of these values may be
    # None.  If both values are None, the frame is removed.
    def setTrackNum(self, n):
+      self.setNum(TRACKNUM_FID, n)
+
+   def setDiscNum(self, n):
+      self.setNum(DISCNUM_FID, n)
+   
+   def setNum(self, fid, n):
       if n[0] == None and n[1] == None:
-         self.frames.removeFramesByID(TRACKNUM_FID);
+         self.frames.removeFramesByID(fid);
          return;
 
       totalStr = "";
@@ -788,9 +797,10 @@ class Tag:
       elif trackStr and not totalStr:
          s = trackStr;
 
-      self.frames.setTextFrame(TRACKNUM_FID, self.strToUnicode(s),
+      self.frames.setTextFrame(fid, self.strToUnicode(s),
                                self.encoding);
 
+
    # Add a comment.  This adds a comment unless one is already present with
    # the same language and description in which case the current value is
    # either changed (cmt != "") or removed (cmt equals "" or None).
@@ -824,8 +834,15 @@ class Tag:
        return self.frames.removeFramesByID(COMMENT_FID);
 
    def addImage(self, type, image_file_path, desc = u""):
-       image_frame = ImageFrame.create(type, image_file_path, desc);
-       self.frames.addFrame(image_frame);
+       if image_file_path:
+           image_frame = ImageFrame.create(type, image_file_path, desc);
+           self.frames.addFrame(image_frame);
+       else:
+           image_frames = self.frames[IMAGE_FID];
+           for i in image_frames:
+               if i.pictureType == type:
+                   self.frames.remove(i);
+                   break;
 
    def getPlayCount(self):
        if self.frames[PLAYCOUNT_FID]:
@@ -867,6 +884,16 @@ class Tag:
       else:
          self.frames.setUniqueFileIDFrame(owner_id, id);
 
+   def getBPM(self):
+      bpm = self.frames[BPM_FID];
+      if bpm:
+          return int(bpm[0].text);
+      else:
+          return None;
+
+   def setBPM(self, bpm):
+       self.setTextFrame(BPM_FID, self.strToUnicode(str(bpm)));
+
    # Test ID3 major version.
    def isV1(self):
       return self.header.majorVersion == 1;
@@ -1238,22 +1265,18 @@ class Tag:
 
 ################################################################################
 class GenreException(Exception):
-   msg = "";
-
-   def __init__(self, msg):
-      self.msg = msg;
-
-   def __str__(self):
-      return self.msg;
-
+   '''Problem looking up genre'''
 
 ################################################################################
 class Genre:
    id = None;
    name = None;
 
-   def __init__(self):
-      pass;
+   def __init__(self, id = None, name = None):
+      if id is not None:
+         self.setId(id);
+      elif name is not None:
+         self.setName(name);
 
    def getId(self):
       return self.id;
@@ -1392,14 +1415,8 @@ class Genre:
       return s;
 
 ################################################################################
-class InvalidAudioFormatException:
-   msg = "";
-
-   def __init__(self, msg):
-      self.msg = msg;
-
-   def __str__(self):
-      return self.msg;
+class InvalidAudioFormatException(Exception):
+   '''Problems with audio format'''
 
 ################################################################################
 class TagFile:
